; boot.asm
; ========
; Author : Jason Cooper
; Description : A boot block that will dump the memory to the boot drive.
;
; Version : 0.2
;   Working fine on test machines
;
; Version : 0.3
;   Refined assembly code to avoid the need for the aditional tool to create the boot block image.
;

org 0x0000
start:
  cli                            ; Switch off interupts as we don't want to be disturbed
  mov    ax, 0x07C0              ; Set up our data segment
  mov    ds, ax

  ; Load the boot block of the first hard disk to make sure that we are from the first hard disk.  This will cause a
  ;   failure unless we are seen by the system as the first hard disk.  (BIOS Depenedent)
  mov   es, ax                  ; Set es to point to point to 0x07C0 so that it will load over our running code
  mov   ax, 0x0201              ; Read 1 sector
  xor   cx,cx                   ; Read from cylinder 0
  xor   bx,bx                   ; Set bx to 0
  xor   dh,dh
  inc   cl                      ; sector 1
  mov   dl, 0x80                ; Read from first head from the first hard disk
  int   0x13                    ; Call the function to read the boot block from the first hard disk.

  ; We should only reach this stage if we are the boot block on the first hard disk.  A small defense from splatting a
  ;   a hard disk but still useful

	; activate FLAT mode
  ;    Basicly it just creates a Global Descripter Table and then tells the CPU to use it.
  ;    After that is done we switch into protected mode and set the ES to use our new descriptor
  ;    Then we switch out of protected mode but the CPU will still let us address the full 32bit range from ES with EBX
  mov [MEM48],0x10
  mov eax,ds
  shl eax,4
  mov bx,GDT
  movzx ebx,bx
  add eax,ebx
  mov dword ptr MEM48+2,eax
  lgdt pword ptr MEM48            ;Load global descriptor table address

  mov bx,0x08                     ;Load bx to point to GDT entry 1
  mov eax,cr0                     ;Switch to protected mode 
  or eax,1
  mov cr0,eax                  
  jmp PROTECTION_ENABLED          ;Clear executionpipe
  PROTECTION_ENABLED:
  mov es,bx
  and al,0xFE                     ;Switch back to real-mode without
  mov cr0,eax                     ;resetting the CPU
  jmp PROTECTION_DISABLED         ;Clear executionpipe
  PROTECTION_DISABLED:
  xor ax,ax                        ; Reset ES to start from 0
  mov es,ax

;  Get Memory Size
  xor   dx,dx                     ; clear dx
  mov    ax, 0xE801
  int    0x15                      ; call function 15/0xE801

;   Note: Some BIOS's return the values in in AX and BX while others put it in 
;         BX and DX

  and    dx,dx                      ; check to see if dx contains 0
  jnz   memoryCont                ; if dx contained a value then good
  mov    dx, bx                     ; else use bx for the upper memory size
memoryCont:
  add    dx, 0x0100                ; add in the lower 16M of memory
  movzx  edx, dx
  shl    edx, 0x10                  ; Convert from blocks of 64k
  mov    [maxMem], edx

; write upto our buffer

  mov  si,DAP                      ; Point si to our DAP
  call  writeBuffer                ; Call the routine to write the buffer out.
  mov    [mem], 0x7E00              ; Store where we have got to
;
; set up pointer to our buffer
;
  xor    ebx, ebx                  ; Clear EBX
  mov    di, 0x7E00                ; Reset di to point to start of our buffer
; Then we hit our main loop of fill the buffer and write it
writeLoop:
  mov    ebx, 0x000b8000            ; Set EBX to point to the first line on the screen
  mov    eax, [mem]                ; Load EAX with where we are through memory at the minute
  xchg  ah,al                      ; Reorder EAX so that we can display it simply with shifts
  ror    eax, 0x10                  ; and calls to our hexDump function
  xchg  ah,al
  mov    cx,  0x04
  mov    dl, 0x1F                  ; Set up the hexDump to be displayed as white on blue

msgLoop:
  call  hexDump                    ;  Call the hexDump with our highest byte
  shr    eax, 8                    ; Move onto the next highest byte
  loop  msgLoop                    ; Keep looping till we have displayed all 4 bytes in EAX

loadBuffer:
  mov    ebx, [mem]                ; Load ebx with the next block of memory to write out
  cmp    ebx, [maxMem]              ;  Compare ebx with the actual amount of memory on the machine
  jge    complete                  ; If it is greater then or equal to we have finished so jmp to complete

  mov    edi, 0x00007E00            ; Reset di to point to start of our buffer
  mov    ecx, 0x00007E00            ; Write 63 sectors
  mov    edx, 0x00007E00
lbLoop:
  mov    eax, [es:ebx]              ;  Load next byte to be moved 
  mov   [es:edx], eax              ;  Store the byte in our buffer
  add    ebx,0x04                  ;  Move onto the next Double Word (i.e. 4 bytes)
  add    edx,0x04                  ;  Move onto the next Double Word (i.e. 4 bytes)
  loop  lbLoop                    ; Keep moving bytes till we have filled our buffer
  mov    [mem], ebx                ; Store where we are up to in moving our memory

  xor    ebx,ebx                    ; Set EBX to 0x00007E00 (our buffer)
  mov    bh,0x7E
  mov    eax, [LBA]                ; Add on the previous secotrs written
  add    eax, 63                    ;
  mov    [LBA],eax
  mov   eax,[LBAH]
  adc    eax,0
  mov    [LBAH],eax
  mov    [buffer],0x00007E00       ; Store the location of our buffer
  call  writeBuffer                ; Call the writeBuffer function

  jmp writeLoop                    ; Jump back to the start of our code for writing out the buffer

;  Hex Dump Function
;    Dump the contents of al (in hex) to the screen location pointed to by ebx
hexDump:
  push  ax
  shr   al, 0x04  ; calculate right digit
  add   al, 0x30  ; convert hex to ascii
  cmp   al, 0x3A  ; is it > 9?
  jl    pIt1  ; jump if digit =0 to 9
  add   al, 0x27  ; digit is A to F
  
pIt1:
  mov    [es:ebx], al
  inc    ebx
  mov    [es:ebx], dl
  inc    ebx

  pop    ax
  and   al, 0x0F  ; mask off left digit
  add   al, 0x30  ; convert hex to ascii
  cmp   al, 0x3A  ; is it > 9?
  jl    pIt2  ; jump if digit =0 to 9
  add   al, 0x27  ; digit is A to F
  
pIt2:
  mov    [es:ebx], al
  inc    ebx
  mov    [es:ebx], dl
  inc    ebx

  ret

; Write Buffer Function
;    Write the buffer out that to the hard disk that is pointed to by ebx
writeBuffer:
  mov    ah, 0x43                  ; Set function to write 1 track
  mov    dl, 0x80                  ;  Write to first hard disk
  int   0x13                      ; Call the function to write the buffer to the disk
  jnc   fine                      ; If fine then jump to fine
  xor    ax,ax                      ; Reset Disk System
  mov    dl,0x80                    ; Reset First Hard Disk
  int    0x13
  inc    [errorCount]              ; Increase error Count
  mov    al, [errorCount]
  cmp    al, 0x04                  ; If four error jump to error Message
  je    error
  jmp writeBuffer
fine:
  mov    [errorCount],0x00          ; Clear error count
  ret

; Error Function
;    Display an error message followed by the error status code and then a memory dump
error:
  mov    [errorStatus], ah          ; Set the error status
  mov    cx, errorLength
  lea    si, [errorMsg]
  call  displayMsg
  mov    al, [errorStatus]
  call  hexDump
  call  memDump
  jmp    endGame

; Memory Dump Function
;   Produce a Hex Dump on the third line of the screen for memory between dumpStart and dumpSize
memDump:
  mov    ebx,0x000b8140
  mov    cx,  dumpSize
  lea    si,  [dumpStart]
  mov    dl,0x1F

dumpLoop:
  lodsb
  call hexDump
  loop dumpLoop
  ret

; Display Message Function
;    Display the message pointed to by ds:si on the second line of the screen
displayMsg:
  mov    ah,  0x1F
  mov    ebx, 0x000b80A0

displayLoop:
  lodsb
  mov    [es:ebx],al
  inc    ebx
  mov    [es:ebx], ah
  inc    ebx
  loop  displayLoop
  ret

;  Completed running
;    Jump to this lump of code when we have succesfully finished
complete:
  mov    cx, completeLength
  lea    si, [completeMsg]
  call  displayMsg
  call   memDump

; End Game Loop
;    Jump here at the end to put the system into a paused state
endGame:
  hlt                  ; Halt the processor
;  jmp endGame


; Data
dataStart:
errorMsg      db  '!'
errorLength   = $-errorMsg

completeMsg    db  '.'
completeLength=  $-completeMsg

              dw  0x0000
dumpStart:
DAP:
DAPHead        db  0x10
              db  0x00
numOfSectors  db  0x3F
              db  0x00
buffer        dd  0x00000000
LBA            dd  0x00000000
LBAH          dd  0x00000000
errorStatus    db  0x00
errorCount    db  0x00
mem            dd  0x00000000
maxMem        dd  0x00000000

dumpSize      = $-dumpStart

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;  global descriptor table (GDT)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
MEM48            db 0x00,0x00,0x00,0x00,0x00,0x00
GDT              db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
                db 0xFF,0xFF,0x00,0x00,0x00,0x92,0xCF,0xFF
gdt_end:
memMap:

times 510-($-$$) db 0

bootBlockId db 0x55, 0xAA
; vim:ts=2 sw=2 nowrap ft=fasm
